"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const opcode_1 = require("./opcode");
const types_1 = require("./types");
const value_1 = require("./value");
const ts_ez_host_1 = require("ts-ez-host");
function gen(code) {
    const op = [];
    const value = [];
    const host = ts_ez_host_1.default();
    const filename = 'mod.tsx';
    host.writeFile(filename, code, false);
    const program = ts.createProgram([filename], {
        jsx: ts.JsxEmit.Preserve,
        experimentalDecorators: true,
        target: ts.ScriptTarget.Latest
    }, host);
    const lexerContext = [];
    const checker = program.getTypeChecker();
    const sourceFile = program.getSourceFile(filename);
    ts.forEachChild(sourceFile, visitor);
    return [op, value];
    function visitor(node) {
        switch (node.kind) {
            case ts.SyntaxKind.NumericLiteral:
                visitNumericLiteral(node);
                break;
            case ts.SyntaxKind.StringLiteral:
                visitStringLiteral(node);
                break;
            case ts.SyntaxKind.BinaryExpression:
                visitBinaryExpression(node);
                break;
            case ts.SyntaxKind.ParenthesizedExpression:
                visitor(node.expression);
                break;
            case ts.SyntaxKind.ConditionalExpression:
                visitConditionalExpression(node);
                break;
            case ts.SyntaxKind.PrefixUnaryExpression:
                visitPrefixUnaryExpression(node);
                break;
            case ts.SyntaxKind.VariableDeclarationList:
                visitVariableDeclarationList(node);
                break;
            case ts.SyntaxKind.WhileStatement:
                visitWhileStatement(node);
                break;
            case ts.SyntaxKind.ForStatement:
                visitForStatement(node);
                break;
            case ts.SyntaxKind.ForInStatement:
            case ts.SyntaxKind.ForOfStatement:
                visitForInOrOfStatement(node);
                break;
            case ts.SyntaxKind.IfStatement:
                visitIfStatement(node);
                break;
            case ts.SyntaxKind.Identifier:
                visitIdentifier(node);
                break;
            case ts.SyntaxKind.Block:
                visitBlock(node);
                break;
            case ts.SyntaxKind.FunctionDeclaration:
            case ts.SyntaxKind.FunctionExpression:
            case ts.SyntaxKind.ArrowFunction:
                visitFunctionLikeDeclaration(node);
                break;
            case ts.SyntaxKind.CallExpression:
                visitCallExpression(node);
                break;
            case ts.SyntaxKind.ReturnStatement:
                visitReturnStatement(node);
                break;
            case ts.SyntaxKind.Parameter:
                visitParameter(node);
                break;
            case ts.SyntaxKind.ElementAccessExpression:
                visitElementAccessExpression(node);
                break;
            case ts.SyntaxKind.PropertyAccessExpression:
                visitPropertyAccessExpression(node);
                break;
            case ts.SyntaxKind.ArrayLiteralExpression:
                visitArrayLiteralExpression(node);
                break;
            case ts.SyntaxKind.ObjectLiteralExpression:
                visitObjectLiteralExpression(node);
                break;
            case ts.SyntaxKind.ThisKeyword:
                visitThisExpression(node);
                break;
            case ts.SyntaxKind.NewExpression:
                visitNewExpression(node);
                break;
            case ts.SyntaxKind.SwitchStatement:
                visitSwitchStatement(node);
                break;
            case ts.SyntaxKind.BreakStatement:
                visitBreakStatement(node);
                break;
            case ts.SyntaxKind.LabeledStatement:
                visitLabeledStatement(node);
                break;
            default:
                ts.forEachChild(node, visitor);
        }
    }
    function createLabel() {
        return { value: op.length };
    }
    function updateLabel(l) {
        l.value = op.length;
    }
    function pushConst(v) {
        value.push(v);
        op.push(opcode_1.OpCode.Const);
        op.push({ value: value.length - 1 });
    }
    function visitIfStatement(stmt) {
        const label1 = createLabel();
        const label2 = createLabel();
        visitor(stmt.expression);
        op.push(opcode_1.OpCode.JumpIfFalse);
        op.push(label2);
        op.push(opcode_1.OpCode.EnterBlockScope);
        visitor(stmt.thenStatement);
        op.push(opcode_1.OpCode.ExitBlockScope);
        op.push(opcode_1.OpCode.Jump);
        op.push(label1);
        updateLabel(label2);
        if (stmt.elseStatement) {
            op.push(opcode_1.OpCode.EnterBlockScope);
            visitor(stmt.elseStatement);
            op.push(opcode_1.OpCode.ExitBlockScope);
        }
        updateLabel(label1);
    }
    function visitLabeledStatement(stmt) {
        const label1 = createLabel();
        pushConst(new value_1.JSString(stmt.label.text));
        op.push(opcode_1.OpCode.EnterLabeledBlockScope);
        op.push(label1);
        visitor(stmt.statement);
        updateLabel(label1);
    }
    function visitBreakStatement(stmt) {
        if (stmt.label) {
            pushConst(new value_1.JSString(stmt.label.text));
            op.push(opcode_1.OpCode.BreakLabel);
        }
        else {
            op.push(opcode_1.OpCode.Break);
        }
    }
    function visitSwitchStatement(stmt) {
        const label1 = createLabel();
        const label2 = createLabel();
        const clauses = stmt.caseBlock.clauses.map(clause => [createLabel(), clause]);
        const defaultClause = clauses.find(([_, clause]) => ts.isDefaultClause(clause));
        visitor(stmt.expression);
        op.push(opcode_1.OpCode.EnterIterableBlockScope);
        op.push(label2);
        clauses.forEach(([label, clause]) => {
            if (!ts.isDefaultClause(clause)) {
                op.push(opcode_1.OpCode.Dup);
                visitor(clause.expression);
                op.push(opcode_1.OpCode.StrictEQ);
                op.push(opcode_1.OpCode.JumpIfTrue);
                op.push(label);
            }
        });
        if (defaultClause) {
            op.push(opcode_1.OpCode.Jump);
            op.push(defaultClause[0]);
        }
        clauses.forEach(([label, clause]) => {
            updateLabel(label);
            clause.statements.forEach(visitor);
        });
        updateLabel(label1);
        op.push(opcode_1.OpCode.ExitBlockScope);
        updateLabel(label2);
    }
    function visitForStatement(stmt) {
        const label1 = createLabel();
        const label2 = createLabel();
        const label3 = createLabel();
        op.push(opcode_1.OpCode.EnterIterableBlockScope);
        op.push(label3);
        stmt.initializer && visitor(stmt.initializer);
        updateLabel(label1);
        if (stmt.condition) {
            visitor(stmt.condition);
        }
        else {
            pushConst(new value_1.JSBoolean(true));
        }
        op.push(opcode_1.OpCode.JumpIfFalse);
        op.push(label2);
        op.push(opcode_1.OpCode.EnterBlockScope);
        visitor(stmt.statement);
        op.push(opcode_1.OpCode.ExitBlockScope);
        stmt.incrementor && visitor(stmt.incrementor);
        op.push(opcode_1.OpCode.Jump);
        op.push(label1);
        updateLabel(label2);
        op.push(opcode_1.OpCode.ExitBlockScope);
        updateLabel(label3);
    }
    function visitForInitializer(initializer) {
        switch (initializer.kind) {
            case ts.SyntaxKind.VariableDeclarationList:
                const declList = initializer;
                declList.declarations.forEach(decl => {
                    pushConst(new value_1.JSString(decl.name.text));
                    switch (getVariableEnvirementType(declList)) {
                        case types_1.EnvironmentType.block:
                            op.push(opcode_1.OpCode.DefBlock);
                            break;
                        default:
                            op.push(opcode_1.OpCode.Def);
                            break;
                    }
                });
        }
    }
    function visitForInOrOfStatement(stmt) {
        const label1 = createLabel();
        const label2 = createLabel();
        const label3 = createLabel();
        visitor(stmt.expression);
        if (ts.isForInStatement(stmt)) {
            op.push(opcode_1.OpCode.ForInStart);
        }
        else {
            op.push(opcode_1.OpCode.ForOfStart);
        }
        op.push(opcode_1.OpCode.EnterIterableBlockScope);
        op.push(label3);
        updateLabel(label1);
        op.push(opcode_1.OpCode.Dup);
        if (ts.isForInStatement(stmt)) {
            op.push(opcode_1.OpCode.ForInNext);
        }
        else {
            op.push(opcode_1.OpCode.ForOfNext);
        }
        op.push(opcode_1.OpCode.JumpIfTrue);
        op.push(label2);
        visitForInitializer(stmt.initializer);
        op.push(opcode_1.OpCode.EnterBlockScope);
        visitor(stmt.statement);
        op.push(opcode_1.OpCode.ExitBlockScope);
        op.push(opcode_1.OpCode.Jump);
        op.push(label1);
        updateLabel(label2);
        op.push(opcode_1.OpCode.ExitBlockScope);
        updateLabel(label3);
    }
    function visitNewExpression(expr) {
        const label1 = createLabel();
        const args = expr.arguments || [];
        args.forEach(visitor);
        op.push(opcode_1.OpCode.Push);
        op.push({ value: args.length });
        visitor(expr.expression);
        op.push(opcode_1.OpCode.New);
        op.push(opcode_1.OpCode.Dup);
        op.push(opcode_1.OpCode.Undefined);
        op.push(opcode_1.OpCode.StrictEQ);
        op.push(opcode_1.OpCode.JumpIfFalse);
        op.push(label1);
        op.push(opcode_1.OpCode.Drop);
        updateLabel(label1);
    }
    function visitThisExpression(expr) {
        op.push(opcode_1.OpCode.This);
    }
    function visitPropertyAccessExpression(propAccess) {
        visitor(propAccess.expression);
        pushConst(new value_1.JSString(propAccess.name.text));
        op.push(opcode_1.OpCode.PropAccess);
    }
    function visitObjectLiteralExpression(obj) {
        obj.properties.forEach(visitObjectLiteralProperty);
        op.push(opcode_1.OpCode.CreateObject);
        op.push({ value: obj.properties.length });
    }
    function visitObjectLiteralProperty(prop) {
        switch (prop.kind) {
            case ts.SyntaxKind.PropertyAssignment: {
                visitor(prop.initializer);
                visitPropertyName(prop.name);
                pushConst(new value_1.JSNumber(types_1.ObjectMemberType.property));
                break;
            }
            case ts.SyntaxKind.GetAccessor: {
                visitFunctionLikeDeclaration(prop);
                visitPropertyName(prop.name);
                pushConst(new value_1.JSNumber(types_1.ObjectMemberType.getter));
                break;
            }
        }
    }
    function visitPropertyName(name) {
        switch (name.kind) {
            case ts.SyntaxKind.Identifier:
                pushConst(new value_1.JSString(name.text));
                break;
            case ts.SyntaxKind.ComputedPropertyName:
            case ts.SyntaxKind.StringLiteral:
            case ts.SyntaxKind.NumericLiteral:
                visitor(name);
                break;
        }
    }
    function visitArrayLiteralExpression(arr) {
        arr.elements.forEach(visitor);
        op.push(opcode_1.OpCode.CreateArray);
        op.push({ value: arr.elements.length });
    }
    function visitElementAccessExpression(elementAccess) {
        visitor(elementAccess.expression);
        visitor(elementAccess.argumentExpression);
        op.push(opcode_1.OpCode.PropAccess);
    }
    function visitParameter(param) {
        if (!ts.isIdentifier(param.name)) {
            throw new Error('not supported');
        }
        pushConst(new value_1.JSString(param.name.text));
        op.push(opcode_1.OpCode.Def);
    }
    function visitReturnStatement(ret) {
        if (ret.expression) {
            visitor(ret.expression);
        }
        else {
            op.push(opcode_1.OpCode.Undefined);
        }
        op.push(opcode_1.OpCode.Ret);
    }
    function visitCallExpression(call) {
        call.arguments.forEach(visitor);
        op.push(opcode_1.OpCode.Push);
        op.push({ value: call.arguments.length });
        visitor(call.expression);
        if (op[op.length - 1] === opcode_1.OpCode.PropAccess) {
            op[op.length - 1] = opcode_1.OpCode.CallMethod;
        }
        else {
            op.push(opcode_1.OpCode.Call);
        }
    }
    function visitFunctionLikeDeclaration(func) {
        const body = func.body;
        if (!body) {
            throw new Error('function must have body');
        }
        const label1 = createLabel();
        const label2 = createLabel();
        op.push(opcode_1.OpCode.Jump);
        op.push(label2);
        updateLabel(label1);
        func.parameters.forEach(visitor);
        lexerContext.push({
            func,
            upValue: new Set()
        });
        if (ts.isBlock(body)) {
            body.statements.forEach(visitor);
            op.push(opcode_1.OpCode.Undefined);
        }
        else {
            visitor(body);
        }
        op.push(opcode_1.OpCode.Ret);
        updateLabel(label2);
        const context = lexerContext.pop();
        context.upValue.forEach(u => pushConst(new value_1.JSString(u)));
        op.push(opcode_1.OpCode.Push);
        op.push({ value: context.upValue.size });
        op.push(opcode_1.OpCode.Push);
        op.push(label1);
        pushConst(new value_1.JSNumber(func.parameters.length));
        if (!ts.isArrowFunction(func)) {
            if (func.name) {
                visitPropertyName(func.name);
            }
            else {
                pushConst(value_1.JSString.Empty);
            }
            op.push(opcode_1.OpCode.CreateFunction);
        }
        else {
            op.push(opcode_1.OpCode.CreateLambda);
        }
    }
    function visitBlock(block) {
        op.push(opcode_1.OpCode.EnterBlockScope);
        block.statements.forEach(visitor);
        op.push(opcode_1.OpCode.ExitBlockScope);
    }
    function visitNumericLiteral(node) {
        pushConst(new value_1.JSNumber(+node.text));
    }
    function visitStringLiteral(node) {
        pushConst(new value_1.JSString(node.text));
    }
    function visitIdentifier(id) {
        pushConst(new value_1.JSString(id.text));
        op.push(opcode_1.OpCode.Load);
        if (lexerContext.length) {
            const context = lexerContext[lexerContext.length - 1];
            if (ts.isArrowFunction(context.func) || id.text !== 'arguments') {
                if (!context.func.locals.has(id.text)) {
                    context.upValue.add(id.text);
                }
            }
        }
    }
    function getVariableEnvirementType(variables) {
        if (variables.flags & ts.NodeFlags.BlockScoped) {
            return types_1.EnvironmentType.block;
        }
        return types_1.EnvironmentType.lexer;
    }
    function visitVariableDeclarationList(variables) {
        variables.declarations.forEach(visitVariableDeclaration.bind(null, getVariableEnvirementType(variables)));
    }
    function visitVariableDeclaration(type, variable) {
        if (!variable.initializer) {
            return;
        }
        visitor(variable.initializer);
        pushConst(new value_1.JSString(variable.name.text));
        switch (type) {
            case types_1.EnvironmentType.block:
                op.push(opcode_1.OpCode.DefBlock);
                break;
            default:
                op.push(opcode_1.OpCode.Def);
                break;
        }
    }
    function visitBinaryExpression(binary) {
        switch (binary.operatorToken.kind) {
            case ts.SyntaxKind.EqualsToken:
            case ts.SyntaxKind.PlusEqualsToken:
                visitAssignmentExpression(binary);
                return;
        }
        visitor(binary.left);
        visitor(binary.right);
        switch (binary.operatorToken.kind) {
            case ts.SyntaxKind.PlusToken:
                op.push(opcode_1.OpCode.Add);
                break;
            case ts.SyntaxKind.MinusToken:
                op.push(opcode_1.OpCode.Sub);
                break;
            case ts.SyntaxKind.AsteriskToken:
                op.push(opcode_1.OpCode.Mul);
                break;
            case ts.SyntaxKind.SlashToken:
                op.push(opcode_1.OpCode.Div);
                break;
            case ts.SyntaxKind.LessThanToken:
                op.push(opcode_1.OpCode.LT);
                break;
            case ts.SyntaxKind.GreaterThanToken:
                op.push(opcode_1.OpCode.GT);
                break;
            case ts.SyntaxKind.EqualsEqualsEqualsToken:
                op.push(opcode_1.OpCode.StrictEQ);
                break;
            case ts.SyntaxKind.ExclamationEqualsEqualsToken:
                op.push(opcode_1.OpCode.StrictNEQ);
                break;
            default:
                throw new Error('not supported');
        }
    }
    function visitAssignmentExpression(expr) {
        switch (expr.operatorToken.kind) {
            case ts.SyntaxKind.EqualsToken: {
                visitor(expr.right);
                visitLeftHandSideExpression(expr.left);
                if (ts.isElementAccessExpression(expr.left) ||
                    ts.isPropertyAccessExpression(expr.left)) {
                    op.push(opcode_1.OpCode.PropAssignment);
                }
                else {
                    op.push(opcode_1.OpCode.Set);
                }
                break;
            }
            case ts.SyntaxKind.PlusEqualsToken:
                {
                    visitor(expr.right);
                    visitor(expr.left);
                    op.push(opcode_1.OpCode.Add);
                    visitLeftHandSideExpression(expr.left);
                    if (ts.isElementAccessExpression(expr.left) ||
                        ts.isPropertyAccessExpression(expr.left)) {
                        op.push(opcode_1.OpCode.PropAssignment);
                    }
                    else {
                        op.push(opcode_1.OpCode.Set);
                    }
                }
                break;
            default:
                throw new Error('not supported');
        }
    }
    function visitWhileStatement(stmt) {
        const label1 = createLabel();
        const label2 = createLabel();
        visitor(stmt.expression);
        op.push(opcode_1.OpCode.JumpIfFalse);
        op.push(label1);
        visitor(stmt.statement);
        op.push(opcode_1.OpCode.Jump);
        op.push(label2);
        updateLabel(label1);
    }
    function visitPrefixUnaryExpression(prefix) {
        switch (prefix.operator) {
            case ts.SyntaxKind.PlusPlusToken: {
                visitor(prefix.operand);
                op.push(opcode_1.OpCode.Push);
                op.push({ value: 1 });
                op.push(opcode_1.OpCode.Add);
                visitLeftHandSideExpression(prefix.operand);
                op.push(opcode_1.OpCode.Set);
                visitor(prefix.operand);
                break;
            }
            default:
                throw new Error('not supported');
        }
    }
    function visitLeftHandSideExpression(lhs) {
        switch (lhs.kind) {
            case ts.SyntaxKind.Identifier:
                pushConst(new value_1.JSString(lhs.text));
                break;
            case ts.SyntaxKind.ElementAccessExpression:
                visitor(lhs.argumentExpression);
                visitor(lhs.expression);
                break;
            case ts.SyntaxKind.PropertyAccessExpression:
                pushConst(new value_1.JSString(lhs.name.text));
                visitor(lhs.expression);
                break;
            default:
                throw new Error('unsupported');
        }
    }
    function visitConditionalExpression(cond) {
        const label1 = createLabel();
        const label2 = createLabel();
        visitor(cond.condition);
        op.push(opcode_1.OpCode.JumpIfFalse);
        op.push(label2);
        visitor(cond.whenTrue);
        op.push(opcode_1.OpCode.Jump);
        op.push(label1);
        updateLabel(label2);
        visitor(cond.whenFalse);
        updateLabel(label1);
    }
}
exports.gen = gen;
//# sourceMappingURL=gen.js.map