import { JSPropertyDescriptor } from './types';
export declare enum ObjectType {
    Number = 0,
    String = 1,
    Boolean = 2,
    Object = 3,
    Function = 4,
    Array = 5,
    Undefined = 6,
    Null = 7,
    ForInOrOfIterator = 8
}
export declare abstract class VObject {
    protected abstract readonly type: ObjectType;
    abstract debugValue(): any;
    isNumber(): this is JSNumber;
    isString(): this is JSString;
    isBoolean(): this is JSBoolean;
    isNull(): this is JSNull;
    isUndefined(): this is JSUndefined;
    isPrimitive(): this is JSPrimitive;
    isObject(): this is JSObject;
    asNumber(): this & JSNumber;
    asString(): this & JSString;
    asBoolean(): JSBoolean;
}
export declare abstract class JSPrimitive extends VObject {
    protected abstract readonly value: any;
    debugValue(): any;
}
export declare class JSUndefined extends JSPrimitive {
    value: undefined;
    readonly type: ObjectType;
    isUndefined(): this is JSUndefined;
    asBoolean(): JSBoolean;
    static instance: JSUndefined;
}
export declare class JSNull extends JSPrimitive {
    value: null;
    readonly type: ObjectType;
    isNull(): this is JSNull;
    asBoolean(): JSBoolean;
    static instance: JSNull;
}
export declare class JSObject extends VObject {
    private rc;
    properties: Map<string | number, JSPropertyDescriptor>;
    constructor(properties?: Map<string | number, VObject>);
    static protoType: JSObject | undefined;
    readonly type: ObjectType;
    debugValue(): any;
    isObject(): this is JSObject;
    isArray(): this is JSArray;
    isFunction(): this is JSFunction;
    asArray(): JSArray;
    get(key: JSString | JSNumber): VObject | JSUndefined;
    set(key: JSString | JSNumber, value: VObject): void;
    getDescriptor(key: JSString | JSNumber): JSPropertyDescriptor | undefined;
    setDescriptor(key: JSString | JSNumber, descriptor: JSPropertyDescriptor): void;
}
export declare class JSNumber extends JSPrimitive {
    value: number;
    constructor(value: number);
    static Zero: JSNumber;
    readonly type: ObjectType;
    asBoolean(): JSBoolean;
}
export declare class JSString extends JSPrimitive {
    value: string;
    constructor(value: string);
    static Empty: JSString;
    readonly type: ObjectType;
}
export declare class JSBoolean extends JSPrimitive {
    value: boolean;
    constructor(value: boolean);
    static False: JSBoolean;
    static True: JSBoolean;
    readonly type: ObjectType;
    asBoolean(): this;
}
export declare class JSFunction extends JSObject {
    name: JSString;
    length: number;
    pos: number;
    upvalue: Map<string, VObject>;
    constructor(name?: JSString, length?: number, pos?: number, upvalue?: Map<string, VObject>);
    static protoType: JSObject;
    isBridge(): this is JSBirdgeFunction;
    isLambda(): this is JSLambda;
    readonly type: ObjectType;
}
export declare class JSLambda extends JSFunction {
    thisObject: VObject;
    length: number;
    pos: number;
    upvalue: Map<string, VObject>;
    constructor(thisObject: VObject, length: number, pos?: number, upvalue?: Map<string, VObject>);
    isLambda(): this is JSLambda;
}
export declare class JSBirdgeFunction extends JSFunction {
    name: JSString;
    func: (this: VObject, ...args: VObject[]) => void;
    constructor(name: JSString, func: (this: VObject, ...args: VObject[]) => void);
    isBridge(): this is JSBirdgeFunction;
    isNative(): this is JSNativeFunction;
    apply(self: VObject, args: VObject[]): void;
}
export declare class JSNativeFunction extends JSBirdgeFunction {
    name: JSString;
    func: (this: VObject, ...args: VObject[]) => VObject;
    constructor(name: JSString, func: (this: VObject, ...args: VObject[]) => VObject);
    isNative(): this is JSNativeFunction;
    apply(self: VObject, args: VObject[]): VObject;
}
export declare class JSArray extends JSObject {
    items: VObject[];
    readonly type: ObjectType;
    constructor(items: VObject[]);
    static protoType: JSObject;
    asArray(): this;
    debugValue(): any[];
}
export declare class JSForInOrOfIterator extends VObject {
    target: JSObject;
    curr: number;
    constructor(target: JSObject);
    readonly type: ObjectType;
    debugValue(): any;
}
