"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("./utils");
const types_1 = require("./types");
var ObjectType;
(function (ObjectType) {
    ObjectType[ObjectType["Number"] = 0] = "Number";
    ObjectType[ObjectType["String"] = 1] = "String";
    ObjectType[ObjectType["Boolean"] = 2] = "Boolean";
    ObjectType[ObjectType["Object"] = 3] = "Object";
    ObjectType[ObjectType["Function"] = 4] = "Function";
    ObjectType[ObjectType["Array"] = 5] = "Array";
    ObjectType[ObjectType["Undefined"] = 6] = "Undefined";
    ObjectType[ObjectType["Null"] = 7] = "Null";
    ObjectType[ObjectType["ForInOrOfIterator"] = 8] = "ForInOrOfIterator";
})(ObjectType = exports.ObjectType || (exports.ObjectType = {}));
class VObject {
    isNumber() {
        return this.type === ObjectType.Number;
    }
    isString() {
        return this.type === ObjectType.String;
    }
    isBoolean() {
        return this.type === ObjectType.Boolean;
    }
    isNull() {
        return false;
    }
    isUndefined() {
        return false;
    }
    isPrimitive() {
        return this.isNumber() || this.isBoolean() || this.isString();
    }
    isObject() {
        return false;
    }
    asNumber() {
        if (this.isNumber()) {
            return this;
        }
        throw new Error('invalid cast');
    }
    asString() {
        if (this.isString()) {
            return this;
        }
        throw new Error('invalid cast');
    }
    asBoolean() {
        throw new Error('invalid cast');
    }
}
exports.VObject = VObject;
class JSPrimitive extends VObject {
    debugValue() {
        return this.value;
    }
}
exports.JSPrimitive = JSPrimitive;
class JSUndefined extends JSPrimitive {
    constructor() {
        super(...arguments);
        this.value = undefined;
    }
    get type() {
        return ObjectType.Undefined;
    }
    isUndefined() {
        return true;
    }
    asBoolean() {
        return new JSBoolean(false);
    }
}
exports.JSUndefined = JSUndefined;
JSUndefined.instance = new JSUndefined();
class JSNull extends JSPrimitive {
    constructor() {
        super(...arguments);
        this.value = null;
    }
    get type() {
        return ObjectType.Null;
    }
    isNull() {
        return true;
    }
    asBoolean() {
        return new JSBoolean(false);
    }
}
exports.JSNull = JSNull;
JSNull.instance = new JSNull();
class JSObject extends VObject {
    constructor(properties = new Map()) {
        super();
        this.rc = 0;
        this.properties = new Map();
        properties.forEach((value, key) => {
            this.properties.set(key, new types_1.JSPropertyDescriptor(value));
        });
        this.setDescriptor(new JSString('__proto__'), new types_1.JSPropertyDescriptor(JSObject.protoType || JSNull.instance, false));
    }
    get type() {
        return ObjectType.Object;
    }
    debugValue() {
        return this;
    }
    isObject() {
        return true;
    }
    isArray() {
        return this.type === ObjectType.Array;
    }
    isFunction() {
        return this.type === ObjectType.Function;
    }
    asArray() {
        throw new Error('invalid cast');
    }
    get(key) {
        if (this.properties.has(key.value)) {
            const descriptor = this.properties.get(key.value);
            return descriptor.value;
        }
        return JSUndefined.instance;
    }
    set(key, value) {
        if (this.properties.has(key.value)) {
            const descriptor = this.properties.get(key.value);
            descriptor.value = value;
            this.setDescriptor(key, descriptor);
        }
        else {
            this.setDescriptor(key, new types_1.JSPropertyDescriptor(value));
        }
    }
    getDescriptor(key) {
        return this.properties.get(key.value);
    }
    setDescriptor(key, descriptor) {
        this.properties.set(key.value, descriptor);
    }
}
exports.JSObject = JSObject;
class JSNumber extends JSPrimitive {
    constructor(value) {
        super();
        this.value = value;
    }
    get type() {
        return ObjectType.Number;
    }
    asBoolean() {
        return new JSBoolean(!!this.value);
    }
}
exports.JSNumber = JSNumber;
JSNumber.Zero = new JSNumber(0);
class JSString extends JSPrimitive {
    constructor(value) {
        super();
        this.value = value;
    }
    get type() {
        return ObjectType.String;
    }
}
exports.JSString = JSString;
JSString.Empty = new JSString('');
class JSBoolean extends JSPrimitive {
    constructor(value) {
        super();
        this.value = value;
    }
    get type() {
        return ObjectType.Boolean;
    }
    asBoolean() {
        return this;
    }
}
exports.JSBoolean = JSBoolean;
JSBoolean.False = new JSBoolean(false);
JSBoolean.True = new JSBoolean(true);
class JSFunction extends JSObject {
    constructor(name = JSString.Empty, length = -1, pos = -1, upvalue = new Map()) {
        super(new Map());
        this.name = name;
        this.length = length;
        this.pos = pos;
        this.upvalue = upvalue;
        this.setDescriptor(new JSString('__proto__'), new types_1.JSPropertyDescriptor(utils_1.assertDef(JSFunction.protoType), false));
        this.setDescriptor(new JSString('prototype'), new types_1.JSPropertyDescriptor(new JSObject(), false));
    }
    isBridge() {
        return false;
    }
    isLambda() {
        return false;
    }
    get type() {
        return ObjectType.Function;
    }
}
exports.JSFunction = JSFunction;
class JSLambda extends JSFunction {
    constructor(thisObject, length, pos = -1, upvalue = new Map()) {
        super(JSString.Empty, length, pos, upvalue);
        this.thisObject = thisObject;
        this.length = length;
        this.pos = pos;
        this.upvalue = upvalue;
    }
    isLambda() {
        return true;
    }
}
exports.JSLambda = JSLambda;
class JSBirdgeFunction extends JSFunction {
    constructor(name, func) {
        super(name);
        this.name = name;
        this.func = func;
    }
    isBridge() {
        return true;
    }
    isNative() {
        return false;
    }
    apply(self, args) {
        this.func.apply(self, args);
    }
}
exports.JSBirdgeFunction = JSBirdgeFunction;
class JSNativeFunction extends JSBirdgeFunction {
    constructor(name, func) {
        super(name, func);
        this.name = name;
        this.func = func;
    }
    isNative() {
        return true;
    }
    apply(self, args) {
        return this.func.apply(self, args);
    }
}
exports.JSNativeFunction = JSNativeFunction;
class JSArray extends JSObject {
    constructor(items) {
        super(new Map(items.map((item, i) => [i, item])));
        this.items = items;
        this.setDescriptor(new JSString('__proto__'), new types_1.JSPropertyDescriptor(utils_1.assertDef(JSArray.protoType), false));
    }
    get type() {
        return ObjectType.Array;
    }
    asArray() {
        return this;
    }
    debugValue() {
        return this.items.map(x => x.debugValue());
    }
}
exports.JSArray = JSArray;
class JSForInOrOfIterator extends VObject {
    constructor(target) {
        super();
        this.target = target;
        this.curr = 0;
    }
    get type() {
        return ObjectType.ForInOrOfIterator;
    }
    debugValue() {
        return this;
    }
}
exports.JSForInOrOfIterator = JSForInOrOfIterator;
//# sourceMappingURL=value.js.map