"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const opcode_1 = require("./opcode");
const utils_1 = require("./utils");
const types_1 = require("./types");
const value_1 = require("./value");
const bom_1 = require("./bom");
class VirtualMachine {
    constructor(codes = [], values = [], cur = 0) {
        this.codes = codes;
        this.values = values;
        this.cur = cur;
        this.stack = [];
        this.frames = [];
        this.environments = [];
        this.initGlobal();
    }
    initGlobal() {
        const valueTable = new Map();
        this.environments.push({
            type: types_1.EnvironmentType.global,
            valueTable
        });
        bom_1.init(this, valueTable);
    }
    popStack() {
        const value = this.stack.pop();
        if (!value) {
            throw new Error('no value');
        }
        return value;
    }
    popCode() {
        const code = this.codes[this.cur++];
        if (!code)
            throw new Error('no code');
        return utils_1.assertOPValue(code);
    }
    lookup(name) {
        for (let i = this.environments.length - 1; i >= 0; i--) {
            const env = this.environments[i];
            if (env.valueTable.has(name)) {
                return env.valueTable.get(name);
            }
            if (env.type === types_1.EnvironmentType.lexer) {
                if (env.upValue.has(name)) {
                    return env.upValue.get(name);
                }
            }
        }
        throw new Error('cannot find name ' + name);
    }
    define(name, value, type) {
        for (let i = this.environments.length - 1; i >= 0; i--) {
            const env = this.environments[i];
            if (type !== types_1.EnvironmentType.block &&
                env.type === types_1.EnvironmentType.block) {
                continue;
            }
            env.valueTable.set(name, value);
            return;
        }
    }
    setValue(name, value) {
        for (let i = this.environments.length - 1; i >= 0; i--) {
            const env = this.environments[i];
            if (env.valueTable.has(name)) {
                env.valueTable.set(name, value);
                return;
            }
            if (env.type === types_1.EnvironmentType.lexer) {
                if (env.upValue.has(name)) {
                    env.upValue.set(name, value);
                    return;
                }
            }
        }
        throw new Error('cannot find name ' + name);
    }
    dump() {
        return {
            stack: this.stack,
            frames: this.frames,
            environments: this.environments,
            codes: this.codes,
            values: this.values,
            cur: this.cur
        };
    }
    load(dump) {
        this.stack = dump.stack;
        this.frames = dump.frames;
        this.environments = dump.environments;
        this.codes = dump.codes;
        this.values = dump.values;
        this.cur = dump.cur;
    }
    step() {
        return this.exec();
    }
    exec(step) {
        while (this.cur < this.codes.length) {
            const op = this.codes[this.cur++];
            switch (op) {
                case opcode_1.OpCode.Const:
                    this.stack.push(this.values[this.popCode()]);
                    break;
                case opcode_1.OpCode.Add: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSNumber(left.asNumber().value + right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.Sub: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSNumber(left.asNumber().value - right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.Mul: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSNumber(left.asNumber().value * right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.Div: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSNumber(left.asNumber().value / right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.LT: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSBoolean(left.asNumber().value < right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.GT: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSBoolean(left.asNumber().value > right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.LTE: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSBoolean(left.asNumber().value <= right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.GTE: {
                    const right = this.popStack();
                    const left = this.popStack();
                    this.stack.push(new value_1.JSBoolean(left.asNumber().value >= right.asNumber().value));
                    break;
                }
                case opcode_1.OpCode.StrictEQ: {
                    const right = this.popStack();
                    const left = this.popStack();
                    if (left.isNumber() && right.isNumber()) {
                        this.stack.push(new value_1.JSBoolean(left.value === right.value));
                    }
                    else {
                        this.stack.push(new value_1.JSBoolean(left === right));
                    }
                    break;
                }
                case opcode_1.OpCode.StrictNEQ: {
                    const right = this.popStack();
                    const left = this.popStack();
                    if (left.isNumber() && right.isNumber()) {
                        this.stack.push(new value_1.JSBoolean(left.value !== right.value));
                    }
                    else {
                        this.stack.push(new value_1.JSBoolean(left !== right));
                    }
                    break;
                }
                case opcode_1.OpCode.Jump: {
                    this.cur = this.popCode();
                    break;
                }
                case opcode_1.OpCode.JumpIfFalse: {
                    const cond = this.popStack();
                    const pos = this.popCode();
                    if (!cond.asBoolean().value) {
                        this.cur = pos;
                    }
                    break;
                }
                case opcode_1.OpCode.JumpIfTrue: {
                    const cond = this.popStack();
                    const pos = this.popCode();
                    if (cond.asBoolean().value) {
                        this.cur = pos;
                    }
                    break;
                }
                case opcode_1.OpCode.Push: {
                    this.stack.push(new value_1.JSNumber(this.popCode()));
                    break;
                }
                case opcode_1.OpCode.Def: {
                    const name = this.popStack();
                    const initializer = this.popStack();
                    this.define(name.asString().value, initializer, types_1.EnvironmentType.lexer);
                    break;
                }
                case opcode_1.OpCode.DefBlock: {
                    const name = this.popStack();
                    const initializer = this.popStack();
                    this.define(name.asString().value, initializer, types_1.EnvironmentType.block);
                    break;
                }
                case opcode_1.OpCode.Load: {
                    const name = this.popStack();
                    const value = this.lookup(name.asString().value);
                    this.stack.push(value);
                    break;
                }
                case opcode_1.OpCode.Set: {
                    const name = this.popStack();
                    const value = this.popStack();
                    this.setValue(name.asString().value, value);
                    break;
                }
                case opcode_1.OpCode.EnterBlockScope: {
                    this.environments.push({
                        type: types_1.EnvironmentType.block,
                        kind: types_1.BlockEnvironmentKind.normal,
                        valueTable: new Map()
                    });
                    break;
                }
                case opcode_1.OpCode.EnterLabeledBlockScope: {
                    const end = this.popCode();
                    const label = this.popStack();
                    this.environments.push({
                        type: types_1.EnvironmentType.block,
                        kind: types_1.BlockEnvironmentKind.labeled,
                        valueTable: new Map(),
                        label: label.asString().value,
                        end
                    });
                    break;
                }
                case opcode_1.OpCode.EnterIterableBlockScope: {
                    const end = this.popCode();
                    this.environments.push({
                        type: types_1.EnvironmentType.block,
                        kind: types_1.BlockEnvironmentKind.iterable,
                        valueTable: new Map(),
                        end
                    });
                    break;
                }
                case opcode_1.OpCode.ExitBlockScope: {
                    this.environments.pop();
                    break;
                }
                case opcode_1.OpCode.Break: {
                    let env = undefined;
                    while (this.environments[this.environments.length - 1].type ===
                        types_1.EnvironmentType.block) {
                        const top = this.environments.pop();
                        if (top.kind === types_1.BlockEnvironmentKind.iterable) {
                            env = top;
                            break;
                        }
                    }
                    if (!env) {
                        throw new Error('cannot break non-iterable block');
                    }
                    this.cur = env.end;
                    break;
                }
                case opcode_1.OpCode.BreakLabel: {
                    let env = undefined;
                    while (this.environments[this.environments.length - 1].type ===
                        types_1.EnvironmentType.block) {
                        const top = this.environments.pop();
                        if (top.kind === types_1.BlockEnvironmentKind.labeled) {
                            env = top;
                            break;
                        }
                    }
                    if (!env) {
                        throw new Error('cannot break non-labeled block');
                    }
                    this.cur = env.end;
                    break;
                }
                case opcode_1.OpCode.CallMethod: {
                    const idx = this.popStack();
                    const obj = this.popStack();
                    if (obj.isObject() && (idx.isNumber() || idx.isString())) {
                        this.getProp(obj, obj, idx);
                    }
                    else {
                        throw new Error('not supported index access');
                    }
                    const callee = this.popStack();
                    const length = this.popStack();
                    if (!callee.isObject() || !callee.isFunction()) {
                        throw new Error('is not callable');
                    }
                    const args = [];
                    for (let i = 0; i < length.asNumber().value; ++i) {
                        args.push(this.popStack());
                    }
                    this.call(callee, args, obj);
                    break;
                }
                case opcode_1.OpCode.Call: {
                    const callee = this.popStack();
                    const length = this.popStack();
                    if (!callee.isObject() || !callee.isFunction()) {
                        throw new Error('is not callable');
                    }
                    const args = [];
                    for (let i = 0; i < length.asNumber().value; ++i) {
                        args.push(this.popStack());
                    }
                    this.call(callee, args, value_1.JSUndefined.instance);
                    break;
                }
                case opcode_1.OpCode.Ret: {
                    const ret = this.popStack();
                    const frame = this.frames.pop();
                    this.cur = frame.ret;
                    this.stack = this.stack.slice(0, frame.entry);
                    this.stack.push(ret);
                    this.environments = this.environments.slice(0, this.environments.indexOf(frame.environments));
                    break;
                }
                case opcode_1.OpCode.PropAccess: {
                    const idx = this.popStack();
                    const obj = this.popStack();
                    if (obj.isObject() && (idx.isNumber() || idx.isString())) {
                        this.getProp(obj, obj, idx);
                    }
                    else {
                        throw new Error('not supported index access');
                    }
                    break;
                }
                case opcode_1.OpCode.ForOfStart:
                case opcode_1.OpCode.ForInStart: {
                    const obj = this.popStack();
                    if (!obj.isObject()) {
                        throw new Error('not supported non-object');
                    }
                    this.stack.push(new value_1.JSForInOrOfIterator(obj));
                    break;
                }
                case opcode_1.OpCode.ForInNext: {
                    const iter = this.popStack();
                    this.forInNext(iter);
                    break;
                }
                case opcode_1.OpCode.ForOfNext: {
                    const iter = this.popStack();
                    this.forOfNext(iter);
                    break;
                }
                case opcode_1.OpCode.CreateArray: {
                    const len = this.popCode();
                    const elements = [];
                    for (let i = 0; i < len; ++i) {
                        elements.push(this.popStack());
                    }
                    const arr = new Array();
                    for (let i = 0; i < len; ++i) {
                        arr.push(elements.pop());
                    }
                    this.stack.push(new value_1.JSArray(arr));
                    break;
                }
                case opcode_1.OpCode.CreateFunction: {
                    const name = this.popStack();
                    const length = this.popStack();
                    const pos = this.popStack();
                    const upValueCount = this.popStack();
                    const upValues = [];
                    const upValue = new Map();
                    for (let i = 0; i < upValueCount.asNumber().value; ++i) {
                        upValues.push(this.popStack().asString().value);
                    }
                    const func = new value_1.JSFunction(name.asString(), length.asNumber().value, pos.asNumber().value, upValue);
                    this.define(name.asString().value, func, types_1.EnvironmentType.lexer);
                    this.stack.push(func);
                    upValues.forEach(name => upValue.set(name, this.lookup(name)));
                    break;
                }
                case opcode_1.OpCode.CreateLambda: {
                    const length = this.popStack();
                    const pos = this.popStack();
                    const upValueCount = this.popStack();
                    const upValues = [];
                    const upValue = new Map();
                    for (let i = 0; i < upValueCount.asNumber().value; ++i) {
                        upValues.push(this.popStack().asString().value);
                    }
                    const func = new value_1.JSLambda(this.frames[this.frames.length - 1].thisObject, length.asNumber().value, pos.asNumber().value, upValue);
                    this.stack.push(func);
                    upValues.forEach(name => upValue.set(name, this.lookup(name)));
                    break;
                }
                case opcode_1.OpCode.CreateObject: {
                    const obj = new value_1.JSObject();
                    const len = this.popCode();
                    for (let i = 0; i < len; ++i) {
                        const type = this.popStack();
                        const name = this.popStack();
                        const initializer = this.popStack();
                        if (type.isNumber() && (name.isString() || name.isNumber())) {
                            switch (type.value) {
                                case types_1.ObjectMemberType.property:
                                    obj.set(name, initializer);
                                    break;
                                case types_1.ObjectMemberType.getter:
                                    if (initializer.isObject() && initializer.isFunction()) {
                                        const descriptor = obj.getDescriptor(name) ||
                                            new types_1.JSPropertyDescriptor(undefined, true, true);
                                        descriptor.getter = initializer;
                                        obj.setDescriptor(name, descriptor);
                                    }
                                    else {
                                        throw new Error('invalid getter');
                                    }
                                    break;
                            }
                        }
                        else {
                            throw new Error('invalid object key');
                        }
                    }
                    this.stack.push(obj);
                    break;
                }
                case opcode_1.OpCode.PropAssignment: {
                    const obj = this.popStack();
                    const idx = this.popStack();
                    const value = this.popStack();
                    if (obj.isObject() && (idx.isNumber() || idx.isString())) {
                        obj.set(idx, value);
                    }
                    else {
                        throw new Error('invalid object assignment');
                    }
                    break;
                }
                case opcode_1.OpCode.This: {
                    this.stack.push(this.frames[this.frames.length - 1].thisObject);
                    break;
                }
                case opcode_1.OpCode.New: {
                    const callee = this.popStack();
                    const length = this.popStack();
                    if (!callee.isObject() || !callee.isFunction()) {
                        throw new Error('is not callable');
                    }
                    const args = [];
                    for (let i = 0; i < length.asNumber().value; ++i) {
                        args.push(this.popStack());
                    }
                    const protoType = callee.get(new value_1.JSString('prototype'));
                    const self = new value_1.JSObject();
                    self.setDescriptor(new value_1.JSString('__proto__'), new types_1.JSPropertyDescriptor(protoType, false));
                    this.stack.push(self);
                    this.call(callee, args, self);
                    break;
                }
                case opcode_1.OpCode.Null: {
                    this.stack.push(value_1.JSNull.instance);
                    break;
                }
                case opcode_1.OpCode.Undefined: {
                    this.stack.push(value_1.JSUndefined.instance);
                    break;
                }
                case opcode_1.OpCode.False: {
                    this.stack.push(value_1.JSBoolean.False);
                    break;
                }
                case opcode_1.OpCode.True: {
                    this.stack.push(value_1.JSBoolean.True);
                    break;
                }
                case opcode_1.OpCode.Zero: {
                    this.stack.push(value_1.JSNumber.Zero);
                    break;
                }
                case opcode_1.OpCode.Drop: {
                    this.stack.pop();
                    break;
                }
                case opcode_1.OpCode.Dup: {
                    const value = this.popStack();
                    this.stack.push(value, value);
                    break;
                }
                case opcode_1.OpCode.Over: {
                    const a = this.popStack();
                    const b = this.popStack();
                    this.stack.push(b, a, b);
                    break;
                }
                case opcode_1.OpCode.Swap: {
                    const a = this.popStack();
                    const b = this.popStack();
                    this.stack.push(a, b);
                    break;
                }
                default:
                    throw new Error('unexpected op: ' + op);
            }
            if (step) {
                return {
                    finished: false
                };
            }
        }
        return {
            finished: true,
            value: this.popStack()
        };
    }
    call(callee, args, thisObject) {
        if (callee.isBridge()) {
            if (callee.isNative()) {
                this.stack.push(callee.apply(thisObject, args.reverse()));
            }
            else {
                callee.apply(thisObject, args.reverse());
            }
            return;
        }
        if (callee.isLambda()) {
            thisObject = callee.thisObject;
        }
        const stackFrame = {
            thisObject,
            ret: this.cur,
            entry: this.stack.length,
            environments: {
                type: types_1.EnvironmentType.lexer,
                valueTable: new Map(),
                upValue: callee.upvalue
            }
        };
        this.frames.push(stackFrame);
        this.environments.push(this.frames[this.frames.length - 1].environments);
        this.cur = callee.pos;
        if (!callee.isLambda()) {
            const argumentsObject = new value_1.JSArray(args);
            argumentsObject.set(new value_1.JSString('callee'), callee);
            this.define('arguments', argumentsObject, types_1.EnvironmentType.lexer);
        }
        args.forEach(x => this.stack.push(x));
        return;
    }
    getProp(obj, curr, key) {
        if (curr.properties.has(key.value)) {
            const descriptor = curr.properties.get(key.value);
            this.getValueByDescriptor(obj, descriptor);
            return;
        }
        const protoType = curr.get(new value_1.JSString('__proto__'));
        if (protoType.isObject()) {
            this.getProp(obj, protoType, key);
            return;
        }
        this.stack.push(value_1.JSUndefined.instance);
    }
    getValueByDescriptor(obj, descriptor) {
        if (descriptor.getter) {
            this.call(descriptor.getter, [], obj);
        }
        else {
            this.stack.push(descriptor.value);
        }
    }
    forInNext(iter) {
        const keys = Array.from(iter.target.properties.entries())
            .filter(([_, value]) => !!value.enumerable)
            .map(([key]) => new value_1.JSString(key.toString()));
        if (iter.curr < keys.length) {
            this.stack.push(keys[iter.curr++]);
            this.stack.push(value_1.JSBoolean.False);
        }
        else {
            this.stack.push(value_1.JSBoolean.True);
        }
    }
    forOfNext(iter) {
        const values = Array.from(iter.target.properties.entries())
            .filter(([_, value]) => !!value.enumerable)
            .map(([_, value]) => value);
        if (iter.curr < values.length) {
            this.getValueByDescriptor(iter.target, values[iter.curr++]);
            this.stack.push(value_1.JSBoolean.False);
        }
        else {
            this.stack.push(value_1.JSBoolean.True);
        }
    }
}
exports.default = VirtualMachine;
//# sourceMappingURL=vm.js.map